/*
 * Copyright (c) 2013, 2017 Oracle and/or its affiliates. All rights reserved. This
 * code is released under a tri EPL/GPL/LGPL license. You can use it,
 * redistribute it and/or modify it under the terms of the:
 *
 * Eclipse Public License version 1.0
 * GNU General Public License version 2
 * GNU Lesser General Public License version 2.1
 */
package org.truffleruby.language.methods;

import com.oracle.truffle.api.CallTarget;
import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
import com.oracle.truffle.api.frame.VirtualFrame;
import com.oracle.truffle.api.object.DynamicObject;
import org.truffleruby.language.RubyNode;
import org.truffleruby.language.Visibility;
import org.truffleruby.language.arguments.RubyArguments;

/**
 * Define a method from a method literal (def mymethod ... end).
 * That is, store the definition of a method and when executed
 * produce the executable object that results.
 */
public class MethodDefinitionNode extends RubyNode {

    private final String name;
    private final SharedMethodInfo sharedMethodInfo;
    private final CallTarget callTarget;

    @Child private GetDefaultDefineeNode getDefaultDefineeNode;

    public MethodDefinitionNode(String name, SharedMethodInfo sharedMethodInfo, CallTarget callTarget) {
        this.name = name;
        this.sharedMethodInfo = sharedMethodInfo;
        this.callTarget = callTarget;
    }

    @TruffleBoundary
    public InternalMethod createMethod(InternalMethod method, Object self, DeclarationContext declarationContext) {
        final DynamicObject dummyModule = coreLibrary().getObjectClass();
        final Visibility dummyVisibility = Visibility.PUBLIC;

        final DynamicObject capturedDefaultDefinee;
        if (declarationContext == DeclarationContext.INSTANCE_EVAL) {
            if (getDefaultDefineeNode == null) {
                getDefaultDefineeNode = insert(new GetDefaultDefineeNode());
            }
            capturedDefaultDefinee = getDefaultDefineeNode.getDefaultDefinee(method, self, declarationContext);
        } else {
            capturedDefaultDefinee = null;
        }

        return new InternalMethod(getContext(), sharedMethodInfo, method.getLexicalScope(), name, dummyModule, dummyVisibility, false, null, callTarget, null, capturedDefaultDefinee);
    }

    @Override
    public Object execute(VirtualFrame frame) {
        return createMethod(RubyArguments.getMethod(frame), RubyArguments.getSelf(frame), RubyArguments.getDeclarationContext(frame));
    }

}
